# Copyright (c) 2009 Dionysus Blazakis <dion@semantiscope.com>. All rights reserved.
#
# See LICENSE.txt for details.

import idaapi
import idc

MARKED_INS_COLOR = 0xffFFa0
MARKED_FUNC_COLOR = DEFCOLOR

marked_funcs = set()

class MarkedFuncsChooser(idaapi.Choose):
  class Item:
    def __init__(self, ea):
      self.ea = ea

    def __str__(self):
      return '%08x' % self.ea

  def __init__(self, list, name):
    idaapi.Choose.__init__(self, list, name)
    self.width = 48

  def getl(self, n):
    if n == 0:
      return "Function Address"
    elif n <= self.sizer():
      return str(self.list[n-1])
    else:
      return "<Empty>"

  def enter(self, n):
    idaapi.jumpto(self.list[n-1].ea)

def mark_func(ea):
  global marked_funcs
  marked_funcs.add(ea)
  idc.SetColor(ea, idc.CIC_FUNC, MARKED_FUNC_COLOR)

def mark_ea(ea):
  idc.SetColor(ea, idc.CIC_ITEM, MARKED_INS_COLOR)

def handle_block(ea):
  curr = ea
  func = idaapi.get_func(ea)
  if func is not None:
    mark_func(func.startEA)

  done = False
  while curr != idaapi.BADADDR:
    if curr != ea:
      xb = idaapi.xrefblk_t()
      ok = xb.first_to(curr, idaapi.XREF_ALL)
      while ok and xb.iscode:
        if xb.type in [idaapi.fl_JF, idaapi.fl_JN]:
          done = True
          break
        ok = xb.next_to()

    if done: break
    mark_ea(curr)
    
    next = idaapi.BADADDR
    xb = idaapi.xrefblk_t()
    ok = xb.first_from(curr, idaapi.XREF_ALL)
    while ok and xb.iscode:
      if xb.type in [idaapi.fl_JF, idaapi.fl_JN]:
        done = True
        break
      elif xb.type == idaapi.fl_F:
        next = xb.to
      ok = xb.next_from()

    if done: break
    curr = next

def main():
  global marked_funcs

  fn = idaapi.askfile_c(0, "results.blocks", "Coverage results file")
  if fn is None:
    idaapi.msg("Aborting...\n")
    return

  base = idaapi.get_imagebase()

  try:
    f = open(fn, "r")
    ea = f.readline()
    while ea != '':
      ea = ea.strip()
      ea = int(ea, 16)
      handle_block(ea)
      ea = f.readline()
  except IOError, ioe:
    idaapi.msg("IOError: [%d] %s\n" % ioe)
  except Exception, e:
    idaapi.msg("Unknown error: %s\n" % e)
  finally:
    f.close()

  mfl = [MarkedFuncsChooser.Item(ea) for ea in marked_funcs]
  c = MarkedFuncsChooser(mfl, "PyIDACov: Covered Functions")
  c.choose()

main()
