#include <stdio.h>
#include <stdint.h>

uint64_t lfsr_parity(uint64_t k)
{
   uint32_t r;

   r = ((k>>32)&0xb0000000) ^ (k&1);
   r = r ^ (r >> 1);
   r = (r ^ (r >> 2)) & 0x11111111;
   r = r*0x11111111;
   return ((r >> 28) & 1);
}

void decrypt(uint32_t *key, unsigned char *data, int len)
{
    uint64_t k, bit;
    int i, cur_bit, c;

    k = ((uint64_t)(key[1])<<32) | key[0];
    for(i = 0; i < len; i++) {
        c = 0;
        for(cur_bit = 7; cur_bit >= 0; cur_bit--) {
/*            printf("%d : %08jx %08jx\n", cur_bit, k>>32, k&0xFFFFFFFF);
            printf("r : %jx\n", ((k>>32)&0xb000) ^ (k&1));*/
            bit = lfsr_parity(k);
            k >>= 1;
            k |= bit<<63;
            c |= (k&1)<<cur_bit;
        }
/*        printf("%08jx %08jx\n", k>>32, k&0xFFFFFFFF);
        printf("c: %x\n", c);*/
        c &= 0xFF;
        data[i] ^= c;
    }
    return;
}

uint64_t data_to_key(const uint8_t *data)
{
    int i=7, b=7;
    uint64_t k = 0, c;
    
    for(i=7; i>= 0; i--) {
        c = 0;
        for(b=0;b<8;b++) {
            c |= (data[i]>>b)&1;
            c <<= 1;
           // printf("%d %d %016jx %016jx\n", i, b, c, k);
        }
        k <<= 8;
     //   printf("%016jx\n", k);
        k ^= c;
    }
    k ^= 1;
    return k;
}

uint64_t rewind_1(uint64_t k)
{
    uint64_t b;
    b = lfsr_parity((k<<1)|(k>>63));
    k = (k<<1)|b;
    return k;
}
uint64_t rewind_n(uint64_t k, int n)
{
    while(n--) {
        k = rewind_1(k);
    }
    return k;
}

int main(int argc, const char *argv[])
{
    unsigned char data[8192];
    uint64_t test = 0;
    uint8_t padd[] = {0x6A, 0xB6, 0x54, 0xC3, 0xCA, 0x8F, 0x53, 0x02};
    int i, n;
    FILE *fin;

    if(argc <2) {
        return 1; 
    }
    fin = fopen(argv[1], "rb");
    fread(data, 1, 8192, fin);
    fclose(fin);


//    decrypt((unsigned int *)key, data, 8192);
// 6724FB980224B2F7
// 0xc930022a au lieu de 4930022a
// 

    n = 8192;
    //padd[0] ^= 0x80;
    test = data_to_key(padd);
    printf("data_to_key : %016jX\n", test); 

    test = rewind_n(test, (n-8)*8);
//    test = 0x98011547225742C0;
    printf("rewind : %016jX\n TO enter : ", test); 
    for(i=4;i<8;i++)
        printf("%02X", (((unsigned char *)&test))[i]);
    for(i=0;i<4;i++)
        printf("%02X", (((unsigned char *)&test))[i]);
    printf("\n");
    decrypt((unsigned int *)&test, data+8192-n, n);
    printf("decrypted : ");
    for(i=8192-n;i<8192;i++)
        printf("%02x", data[i]);
    printf("\n");


    fin = fopen("fuuu", "wb");
    fwrite(data, 1, 8192, fin);
    fclose(fin);
    return 0;
}
